@@ -1,4 +1,4 @@
-/* Copyright 1995-1998,2000-2014,2018-2019
+/* Copyright 1995-1998,2000-2014,2018-2019,2023-2025
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -32,6 +32,7 @@
 #include <stdlib.h>
 #include <string.h>   /* for memset used by FD_ZERO on Solaris 10 */
 #include <sys/time.h>
+#include <sys/select.h>                           /* from Gnulib */
 #include <unistd.h>
 
 #if HAVE_PTHREAD_NP_H
@@ -46,6 +47,7 @@
 #include "dynwind.h"
 #include "eval.h"
 #include "extensions.h"
+#include "finalizers.h"
 #include "fluids.h"
 #include "gc-inline.h"
 #include "gc.h"
@@ -354,6 +356,9 @@ scm_i_pthread_key_t scm_i_thread_key;
    itself in TLS (rather than a pointer to some malloc'd memory) is not
    possible since thread objects may live longer than the actual thread they
    represent.  */
+# ifdef __APPLE__ /* XXX: hack to address <https://bugs.gnu.org/60234> */
+SCM_INTERNAL
+# endif
 SCM_THREAD_LOCAL scm_thread *scm_i_current_thread = NULL;
 
 #endif /* SCM_HAVE_THREAD_STORAGE_CLASS */
@@ -363,7 +368,7 @@ static scm_i_pthread_mutex_t thread_admin_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZE
 static scm_thread *all_threads = NULL;
 static int thread_count;
 
-static SCM default_dynamic_state;
+static SCM default_dynamic_state = SCM_BOOL_F;
 
 /* Perform first stage of thread initialisation, in non-guile mode.
  */
@@ -488,9 +493,8 @@ on_thread_exit (void *v)
   t->handle = SCM_PACK (0);
 
   /* If there's only one other thread, it could be the signal delivery
-     thread, so we need to notify it to shut down by closing its read pipe.
-     If it's not the signal delivery thread, then closing the read pipe isn't
-     going to hurt.  */
+     thread, in which case we should shut it down also by closing its
+     read pipe.  */
   if (thread_count <= 1)
     scm_i_close_signal_pipe ();
 
@@ -591,6 +595,11 @@ scm_i_init_thread_for_guile (struct GC_stack_base *base,
             needs_unregister = 1;
 #endif
 
+          // We were blocked while another thread was initializing
+          // guile, so we need to re-read the (now correct) state.
+          if (scm_is_eq (dynamic_state, SCM_BOOL_F))
+            dynamic_state = default_dynamic_state;
+
 	  guilify_self_1 (base, needs_unregister);
 	  guilify_self_2 (dynamic_state);
 	}
@@ -898,7 +907,7 @@ enum scm_mutex_kind {
   /* An unowned mutex is like a standard mutex, except that it can be
      unlocked by any thread.  A corrolary of this behavior is that a
      thread's attempt to lock a mutex that it already owns will block
-     instead of signalling an error, as it could be that some other
+     instead of signaling an error, as it could be that some other
      thread unlocks the mutex, allowing the owner thread to proceed.
      This kind of mutex is a bit strange and is here for use by
      SRFI-18.  */
@@ -1022,14 +1031,7 @@ lock_mutex (enum scm_mutex_kind kind, struct scm_mutex *m,
 
         if (err == 0)
           {
-            if (scm_is_eq (m->owner, SCM_BOOL_F))
-              {
-                m->owner = current_thread->handle;
-                scm_i_pthread_mutex_unlock (&m->lock);
-                return SCM_BOOL_T;
-              }
-            else
-              continue;
+            goto maybe_acquire;
           }
         else if (err == ETIMEDOUT)
           {
@@ -1041,7 +1043,7 @@ lock_mutex (enum scm_mutex_kind kind, struct scm_mutex *m,
             scm_i_pthread_mutex_unlock (&m->lock);
             scm_async_tick ();
             scm_i_scm_pthread_mutex_lock (&m->lock);
-            continue;
+            goto maybe_acquire;
           }
         else
           {
@@ -1050,6 +1052,14 @@ lock_mutex (enum scm_mutex_kind kind, struct scm_mutex *m,
             errno = err;
             SCM_SYSERROR;
           }
+
+      maybe_acquire:
+        if (scm_is_eq (m->owner, SCM_BOOL_F))
+          {
+            m->owner = current_thread->handle;
+            scm_i_pthread_mutex_unlock (&m->lock);
+            return SCM_BOOL_T;
+          }
       }
 }
 #undef FUNC_NAME
@@ -1165,7 +1175,7 @@ SCM_DEFINE (scm_unlock_mutex, "unlock-mutex", 1, 0, 0, (SCM mutex),
             "Unlocks @var{mutex}.  The calling thread must already hold\n"
             "the lock on @var{mutex}, unless the mutex was created with\n"
             "the @code{allow-external-unlock} option; otherwise an error\n"
-            "will be signalled.")
+            "will be signaled.")
 #define FUNC_NAME s_scm_unlock_mutex
 {
   struct scm_mutex *m;
@@ -1387,15 +1397,16 @@ timed_wait (enum scm_mutex_kind kind, struct scm_mutex *m, struct scm_cond *c,
 
 SCM_DEFINE (scm_timed_wait_condition_variable, "wait-condition-variable", 2, 1, 0,
 	    (SCM cond, SCM mutex, SCM timeout),
-"Wait until condition variable @var{cv} has been signalled.  While waiting, "
+"Wait until condition variable @var{cv} has been signaled.  While waiting, "
 "mutex @var{mx} is atomically unlocked (as with @code{unlock-mutex}) and "
 "is locked again when this function returns.  When @var{t} is given, "
 "it specifies a point in time where the waiting should be aborted.  It "
 "can be either a integer as returned by @code{current-time} or a pair "
 "as returned by @code{gettimeofday}.  When the waiting is aborted the "
-"mutex is locked and @code{#f} is returned.  When the condition "
-"variable is in fact signalled, the mutex is also locked and @code{#t} "
-"is returned. ")
+"mutex is locked and @code{#f} is returned.  After the condition "
+"variable is signaled, the mutex is locked and @code{#t} is returned.  "
+"@code{#t} may also be returned spuriously, so any relevant conditions "
+"should be re-checked.")
 #define FUNC_NAME s_scm_timed_wait_condition_variable
 {
   scm_t_timespec waittime_val, *waittime = NULL;
@@ -1676,18 +1687,19 @@ SCM_DEFINE (scm_all_threads, "all-threads", 0, 0, 0,
 	    "Return a list of all threads.")
 #define FUNC_NAME s_scm_all_threads
 {
-  /* We can not allocate while holding the thread_admin_mutex because
-     of the way GC is done.
-  */
-  int n = thread_count;
   scm_thread *t;
-  SCM list = scm_c_make_list (n, SCM_UNSPECIFIED), *l;
 
   scm_i_pthread_mutex_lock (&thread_admin_mutex);
-  l = &list;
+
+  int n = thread_count;
+  SCM list = scm_c_make_list (n, SCM_UNSPECIFIED);
+  SCM *l = &list;
+
   for (t = all_threads; t && n > 0; t = t->next_thread)
     {
-      if (t != scm_i_signal_delivery_thread)
+      if (!t->exited
+          && !scm_i_is_signal_delivery_thread (t)
+          && !scm_i_is_finalizer_thread (t))
 	{
 	  SCM_SETCAR (*l, t->handle);
 	  l = SCM_CDRLOC (*l);
@@ -1695,7 +1707,9 @@ SCM_DEFINE (scm_all_threads, "all-threads", 0, 0, 0,
       n--;
     }
   *l = SCM_EOL;
+
   scm_i_pthread_mutex_unlock (&thread_admin_mutex);
+
   return list;
 }
 #undef FUNC_NAME
@@ -1818,7 +1832,6 @@ scm_init_threads ()
 					 sizeof (struct scm_cond));
   scm_set_smob_print (scm_tc16_condvar, scm_cond_print);
 
-  default_dynamic_state = SCM_BOOL_F;
   guilify_self_2 (scm_i_make_initial_dynamic_state ());
   threads_initialized_p = 1;
 
